--[[ SlashPrint
     A /print command to evaluate lua expressions

     Also a dump function that drops things into a table.

]]--

local addoninfo, SlashPrint = ...

SlashPrint.aborted = false
SlashPrint.maxlines = 10000
SlashPrint.maxdepth = 10
SlashPrint.maxitems = 1000

local printf = Library.printf.printf
local printfh = Library.printf.printfh
local printfs = Library.printf.printfs
local sprintf = Library.printf.sprintf
local sprintfh = Library.printf.sprintfh

local grey = '<font color="#aaaaaa">'
local name_color = '<font color="#ccccee">'
local normal_color = '<font color="#eeeeee">'
local special_color = '<font color="#ddeedd">'

function SlashPrint.append(tab, fmt, ...)
  tab[#tab + 1] = sprintfh(fmt, ...)
end

function SlashPrint.stringify(val, tablekey)
  local t = type(val)
  local pretty
  local color = normal_color
  if     t == 'table' then
    pretty = tostring(val)
    color = special_color
  elseif t == 'string' then
    if tablekey and string.match(val, '^[_%a][_%a%d]*$') then
      pretty = string.format("%s", val)
      color = name_color
    else
      pretty = string.format("\"%s\"", val)
    end
  elseif t == 'function' then
    pretty = tostring(val)
    color = special_color
  elseif t == 'boolean' then
    pretty = val and "true" or "false"
  elseif t == 'number' then
    if (math.floor(val) == val) then
      pretty = string.format("%d", val)
    else
      pretty = string.format("%f", val)
    end
  elseif t == 'userdata' then
    pretty = tostring(val)
    color = special_color
  elseif t == 'thread' then
    pretty = tostring(val)
    color = special_color
  elseif t == 'nil' then
    pretty = "nil"
  else
    pretty = "<unknown type " .. t .. ">"
  end
  return pretty, color
end

function SlashPrint.empty(tab)
  for k, v in pairs(tab) do
    if v ~= nil or SlashPrint.verbose then
      return false
    end
  end
  return true
end

function SlashPrint.dump(tab, val, indent, comma, html)
  local t = type(val)
  if not indent then
    SlashPrint.visited = {}
    indent = 0
  end
  local ifmt = string.format("%%%ds", (indent + 1) * 2)
  local istr = string.format(ifmt, "")
  local cstr = comma and "," or ""
  if SlashPrint.aborted then
    if t == 'table' then
      SlashPrint.visited[val] = true
    end
    return
  end
  if #tab >= SlashPrint.maxlines then
    SlashPrint.append(tab, "... Too many lines, stopping.")
    SlashPrint.aborted = true
    return
  end
  if      t == 'table' then
    SlashPrint.visited[val] = true
    -- and let the trailing bit handle this

    if SlashPrint.empty(val) then
      if html then
        SlashPrint.append(tab, grey .. "%s  {}%s</font>", istr, cstr)
      else
        SlashPrint.append(tab, "%s  {}%s", istr, cstr)
      end
      return
    end
    if indent < 1 then
      if html then
        SlashPrint.append(tab, "%s" .. grey .. "{</font>", istr)
      else
        SlashPrint.append(tab, "%s{", istr)
      end
    end
    local counter = 1
    for k, v in pairs(val) do
      if counter > SlashPrint.maxitems then
	if html then
	  SlashPrint.append(tab, grey .. "%s  ...</font>", istr);
	else
	  SlashPrint.append(tab, "%s  ...", istr);
	end
        break
      end
      counter = counter + 1
      if SlashPrint.aborted then
        break
      end
      if #tab >= SlashPrint.maxlines then
	SlashPrint.append(tab, "... Too many lines, stopping.")
	SlashPrint.aborted = true
        break
      end
      pretty_k, key_color = SlashPrint.stringify(k, true)
      pretty_v, value_color = SlashPrint.stringify(v, false)
      if type(v) == 'table' then
	if SlashPrint.visited[v] then
	  if html then
            SlashPrint.append(tab, "%s  " .. key_color .. "%s</font> " .. grey .. "= {</font> " .. value_color .. "%s</font> " .. grey .. " (already visited) }</font>", istr, pretty_k, pretty_v)
	  else
            SlashPrint.append(tab, "%s  %s = { %s (already visited) }", istr, pretty_k, pretty_v)
	  end
	elseif indent + 1 >= SlashPrint.maxdepth then
	  if html then
            SlashPrint.append(tab, "%s  " .. key_color .. "%s</font> " .. grey .. "= {</font> " .. value_color .. "%s</font> " .. grey .. " (depth limit reached) }</font>", istr, pretty_k, pretty_v)
	  else
            SlashPrint.append(tab, "%s  %s = { %s (depth limit reached) }", istr, pretty_k, pretty_v)
	  end
	else
	  if SlashPrint.empty(v) then
	    if html then
              SlashPrint.append(tab, "%s  " .. key_color .. "%s</font> " .. grey .. "= {},</font>", istr, pretty_k)
	    else
              SlashPrint.append(tab, "%s  %s = {},", istr, pretty_k)
	    end
	  else
	    if html then
              SlashPrint.append(tab, "%s  " .. key_color .. "%s</font> " .. grey .. "= " .. value_color .. "%s</font> " .. grey .. "{</font>", istr, pretty_k, pretty_v)
	    else
              SlashPrint.append(tab, "%s  %s = %s {", istr, pretty_k, pretty_v)
	    end
	    SlashPrint.dump(tab, v, indent + 1, true, html)
	    if html then
	      SlashPrint.append(tab, "%s  " .. grey .. "},</font>", istr)
	    else
	      SlashPrint.append(tab, "%s  },", istr)
	    end
	  end
	end
      else
        if v ~= nil or SlashPrint.verbose then
          pretty_v = SlashPrint.stringify(v)
	  if html then
	    SlashPrint.append(tab, "%s  " .. key_color .. "%s</font> " .. grey .. "=</font> " .. value_color .. "%s</font>" .. grey .. ",</font>", istr, pretty_k, pretty_v)
	  else
	    SlashPrint.append(tab, "%s  %s = %s,", istr, pretty_k, pretty_v)
	  end
        end
      end
    end
    if indent < 1 then
      if html then
	SlashPrint.append(tab, grey .. "%s}%s</font>", istr, cstr)
      else
	SlashPrint.append(tab, "%s}%s", istr, cstr)
      end
    end
  else
    local pretty = SlashPrint.stringify(val)
    if html then
      SlashPrint.append(tab, "%s%s" .. grey .. "%s</font>", istr, pretty, cstr)
    else
      SlashPrint.append(tab, "%s%s%s", istr, pretty, cstr)
    end
  end
end

function SlashPrint.slashcommand(args)
  SlashPrint.aborted = false
  if not args then
    printf("Usage error.")
    return
  end
  SlashPrint.maxdepth = args.d or 10
  SlashPrint.maxitems = args.m or 1000
  SlashPrint.verbose = args.v
  local func, error = loadstring("return { " .. args.leftover_full .. " }")
  SlashPrint.visited = {}
  if func then
    local status, val = pcall(func)
    local pretty
    if status then
      printf("Evaluated '%s':", args['leftover'])
      SlashPrint.visited[val] = true
      if #val == 0 then
        printf("Got no values returned (or nil).")
      elseif #val == 1 then
	-- we only got one arg
	pretty = {}
        SlashPrint.dump(pretty, val[1], false, false, true)
	for i = 1, #pretty do
	  Command.Console.Display('general', true, pretty[i], true)
	  if i > 1000 then
	    printfh(grey .. '... %d more lines omitted.</font>',
	      #pretty - 1001)
	    break
	  end
	end
      else
        printf("Got %d results:", #val)
        for i, v in ipairs(val) do
	  if SlashPrint.aborted then
	    break
	  end
	  pretty = {}
          SlashPrint.dump(pretty, v, false, false, true)
	  for idx, v in ipairs(pretty) do
	    Command.Console.Display('general', true, v, true)
	    if idx > 1000 then
	      printfh(grey .. '... %d more lines omitted.</font>',
	      	#pretty - 1001)
	      break
	    end
	  end
        end
      end
      if SlashPrint.aborted then
        printf("Stopped after %d lines.", #pretty)
      end
    else
      printf("Error evaluating <%s>: %s", args['leftover'], val)
    end
  else
    printf("Couldn't load <%s>: %s", args['leftover'], error)
  end
end

Library.LibGetOpt.makeslash("d#m#v", "SlashPrint", "print", SlashPrint.slashcommand)
